import java.util.*;

@SuppressWarnings("serial")
public class Network extends HashMap<String, Node> {

	public Network() {
	}

	public Network(List<String> linelist) {
		init(linelist);
	}

	public void init(List<String> linelist) {
		for (String s : linelist) {
			if (s.toUpperCase().startsWith("P")) {
				insertProbability(s);
			}
		}
		connectNodes();
	}

	private void connectNodes() {
		Collection<Node> nodes = this.values();
		for (Node node : nodes) {
			for (String parentName : node.parentsName) {
				node.parents.add(this.get(parentName));
			}
		}
	}

	public void printNodes() {
		for (String s : keySet()) {
			Node n = get(s);
			System.out.println(s+":");
			System.out.println(n.table.toString());
		}
	}

	private double bayes(List<Identifier> vars) {
		HashSet<Identifier> conditions = new HashSet<Identifier>();
		double v = 1.0;
		// Loop through all nodes values.
		for (Identifier id : vars) {
			boolean hasParents = true;
			conditions.clear();
			Node n = this.get(id.getId());
			// Check parent values
			for (Node p : n.parents) {
				for (Identifier i : vars) {
					if (i.getId().equals(p.id)) {
						conditions.add(i);
						break;
					}
				}
			}
			// If node has no parents, we get value of ourself.
			if (conditions.isEmpty()) {
				conditions.add(id);
				hasParents = false;
			}
			double d = n.getValue(conditions);
			// If node has parents and is false, we need to get the false value
			if (hasParents && (id.getValue() == false)) {
					d = 1.0-d;
			}
			v *= d; 
		}
		return v;
	}

	@SuppressWarnings("unchecked")
	public double enumerateAll(ArrayList<Identifier> known,
			ArrayList<Identifier> unknown) {

		if (unknown.size() == 0) {
			// We now know all nodes values, retrieve probability.
			return bayes(known);
		}
		// Move one node from unknown to known, and iterate over true/false.
		Identifier a = unknown.remove(0);
		known.add(a);
		ArrayList<Identifier> un1 = (ArrayList<Identifier>)unknown.clone();
		ArrayList<Identifier> un2 = (ArrayList<Identifier>)unknown.clone();
		ArrayList<Identifier> k1 = (ArrayList<Identifier>)known.clone();
		ArrayList<Identifier> k2 = (ArrayList<Identifier>)known.clone();
		double v = enumerateAll(k1, un1);
		a.toggleValue();
		v += enumerateAll(k2, un2);
		// Return the sum of the true/false probabilities.
		return v;
	}

	/**
	 * Creates a Identifier from a string. For example "!Symptom" will become
	 * an Identifier with id "Symptom" and value false.
	 * @param name String representation of a node value.
	 * @return Itentifier representation of the same.
	 */
	public static Identifier stringToIdentifier(String name) {
		String id;
		boolean value;
		if (name.startsWith("!")) {
			id = name.substring(1);
			value = false;
		} else {
			id = name;
			value = true;
		}
		return new Identifier(id, value);
	}

	public void insertProbability(String s) {
		String[] sides = s.split("=");
		double d = Double.parseDouble(sides[1]);
		String s2 = sides[0].substring(sides[0].indexOf('(')+1, sides[0].indexOf(')'));
		String sides2[] = s2.split("\\|");
		String id;
		if (sides2[0].startsWith("!")) {
			id = sides2[0].substring(1);
		} else {
			id = sides2[0];
		}
		Node n = this.get(id);
		if (n == null) {
			n = new Node(id);
			this.put(id, n);
		}
		if (sides2.length == 1) {
			Identifier i1, i2;
			if (sides2[0].startsWith("!")) {
				i2 = new Identifier(id, true);
				i1 = new Identifier(sides2[0].substring(1), false);	
			} else {
				i1 = new Identifier(id, true);
				i2 = new Identifier(sides2[0], false);
			}
			
			HashSet<Identifier> hs1 = new HashSet<Identifier>();
			HashSet<Identifier> hs2 = new HashSet<Identifier>();
			hs1.add(i1);
			hs2.add(i2);
			n.table.put(hs1, d);
			n.table.put(hs2, 1.0-d);
		} else {
			String[] parts3 = sides2[1].split(",");
			HashSet<Identifier> hs = new HashSet<Identifier>();
			// Add parents names if they are not already there.
			boolean parentsCreated = true;
			if (n.parentsName.isEmpty()) {
				parentsCreated = false;
			}
			for (String s3 : parts3) {
				Identifier i;
				if (s3.startsWith("!")) {
					i = new Identifier(s3.substring(1), false);
				} else {
					i = new Identifier(s3, true);
				}
				hs.add(i);
				if (!parentsCreated) {
					n.parentsName.add(i.getId());
				}
			}
			n.table.put(hs, d);
		}
	}

	@SuppressWarnings("unchecked")
	public double Prob(List<Identifier> asks, ArrayList<Identifier> cond) {
		cond.addAll(asks);
		ArrayList<Identifier> unknown = new ArrayList<Identifier>();
		for (String id : this.keySet()) {
			boolean found = false;
			for (Identifier var : cond) {
				if (id.equals(var.getId())) {
					found = true;
					break;
				}
			}
			if (!found) {
				unknown.add(new Identifier(id, true));
			}
		}
		ArrayList<Identifier> unknown2 = (ArrayList<Identifier>) unknown.clone();
		ArrayList<Identifier> cond2 = (ArrayList<Identifier>) cond.clone();

		double[] dist = new double[2];
		dist[0] = enumerateAll(cond, unknown);
		for (Identifier id : asks) {
			id.toggleValue();
		}
		dist[1] = enumerateAll(cond2, unknown2);
		dist = normalize(dist);
		return dist[0];
	}
	
	public double[] normalize(double[] dist) {
		int len = dist.length;
		double total = dist[0] + dist[1];
		double[] normalized = new double[len];
		if (total != 0) {
			for (int i = 0; i < len; i++) {
				normalized[i] = dist[i] / total;
			}
		}
		return normalized;
	}
}
